;; Problem 24
;; 16 August 2002
;;
;; A permutation is an ordered arrangement of objects. For example,
;; 3124 is one possible permutation of the digits 1, 2, 3 and 4. If
;; all of the permutations are listed numerically or alphabetically,
;; we call it lexicographic order. The lexicographic permutations of
;; 0, 1 and 2 are:
;;
;; 012   021   102   120   201   210
;;
;; What is the millionth lexicographic permutation of the digits 0, 1, 2, 3, 4, 5, 6, 7, 8 and 9?

;; if we fix the 0 in the first slot, we have (fact 9) perms left
;;   (fact 9) 362880
;; if we fix the 1, we go up to
;;   (* 2 362880) 725760
;; if we fix the 2, we go up to
;;   (* 3 362880) 1088640
;; so the first num must be 3
;; fixing the 3 then represents the 725761'st permutation
;; fixing the 0 as the next gives us (fact 8) 40320 more
;; (- 1000000 725761) 274239
;; (floor (/ 274239 40320)) ; 6
;; (* 6 40320) 241920
;; (+ 725761 241920) ; 967681
;; (- 1000000 967682) ;; 32318
;; take the 6th thing from whats left: (subseq "012456789" 6 7) "7"


(defun find-nth-perm (fixed things nth)
  (format t "~a ~a ~a~&" fixed things nth)
  (if (= 1 nth)
      (return-from find-nth-perm
	(append (reverse fixed) things)))
  (let* ((num-perms-eliminated-by-fixing-first (fact (- (length things) 1)))
	 (mult (loop for ii = 0 then (+ ii 1)
		     while (< num-perms-eliminated-by-fixing-first nth)
		     finally (return ii)
		     do
		     (setf nth (- nth num-perms-eliminated-by-fixing-first)))))
    (find-nth-perm (cons (nth mult things) fixed)
		   (remove (nth mult things) things)
		   nth)))

; (find-nth-perm '() '(0 1 2) 6)
; (find-nth-perm '() '(0 1 2 3) 3)
(find-nth-perm '() '(0 1 2 3 4 5 6 7 8 9) 1000000)

;; NIL (0 1 2 3 4 5 6 7 8 9) 1000000
;; (2) (0 1 3 4 5 6 7 8 9) 274240
;; (7 2) (0 1 3 4 5 6 8 9) 32320
;; (8 7 2) (0 1 3 4 5 6 9) 2080
;; (3 8 7 2) (0 1 4 5 6 9) 640
;; (9 3 8 7 2) (0 1 4 5 6) 40
;; (1 9 3 8 7 2) (0 4 5 6) 16
;; (5 1 9 3 8 7 2) (0 4 6) 4
;; (4 5 1 9 3 8 7 2) (0 6) 2
;; (6 4 5 1 9 3 8 7 2) (0) 1
;; (2 7 8 3 9 1 5 4 6 0)

;; (2783915460)



(defun fact (num)
  (declare (optimize speed)
	   (type fixnum num))
  (cond ((= 0 num)
	 1)
	(t
	 (loop 
	  for ii = num then (- ii 1)
	  for total = (* 1 ii) then (* total ii)
	  do
	  (when (= 1 ii)
	    (return total))))))

;(fact 0)

(defun n-choose-k (nn kk)
  (declare (optimize speed)
	   (type fixnum nn kk))
  (/
   (fact nn)
   (*
    (fact kk)
    (fact (- nn kk)))))

;; (n-choose-k 4 2)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; second attempt


;; ;; the number of ways to arrange the 10 items is 10!
;; ;; or: (fact 10) = 3628800
;; ;; the number of ways to arrange 0 . (rest) is
;; ;;     (fact 9) = 362880
;; ;; so the first number is not 0, 2 takes us to 
;; ;;   (* 2 362880) 725760
;; ;; 3 starts at 725761 and takes us up to 
;; ;;   (* 3 362880) 1088640
;; ;; so it must start w/3
;; ;; (fact 8) 40320

;; bad: 3671804259
;; ;; number: 2671804359
;; ;; rest: 359

;; ;(- 1000000 725761) ; 274239
;; ;;(floor (/ 274239 40320))
;; ; (* 6 40320); 241920

;; (+ 725761 241920); 967681
;; (- 1000000 967681); 32319
;; ; (fact 7); 5040
;; (* 6 5040); 30240
;; (- 32319 30240); 2079
;; (fact 6); 720
;; (* 2 720); 1440
;; (- 2079 1440); 639
;; (fact 5); 120
;; (* 5 120); 600
;; (- 639 600); 39
;; (fact 4); 24

;; (- 39 24); 15
;; (fact 3); 6
;; (- 15 12); 3
;; (fact 2); 2

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; first attempt...
;; (defun curry (fn &rest outer-args)
;;   (lambda (&rest args)
;;     (apply fn (append outer-args args))))

;; (defun rcurry (fn &rest outer-args)
;;   (lambda (&rest args)
;;     (apply fn (append args outer-args))))

;; ;; (let ((test (curry #'equalp "a")))
;; ;;   (format t "a:~a~&" (funcall test "a"))
;; ;;   (format t "b:~a~&" (funcall test "b")))

;; (defun flatten (tree)
;;   (labels ((inner-flatten (acc tree)
;; 	     (format t "inner-flatten, acc:~a tree:~a~&" acc tree)
;; 	     (cond ((null tree)
;; 		    acc)
;; 		   ((listp (first tree))
;;  		    (inner-flatten
;;  		     (inner-flatten acc (first tree))
;;  		     (rest tree)))
;; 		   (t
;; 		    (inner-flatten
;; 		     (cons (first tree) acc)
;; 		     (rest tree))))))
;;     (reverse (inner-flatten (list) tree))))

;; ; (flatten '(1 ((2 3) 4)))

;; (defun all-permutations (items)
;;   (cond ((= 1 (length items))
;; 	 (list items))
;; 	(t
;; 	 (let ((result (list)))
;; 	   (loop for item in items
;; 		 do
;; 		 (let ((remainder (remove-if (curry #'equalp item) items)))
;; 		   (loop for perm in (all-permutations remainder)
;; 			 do
;; 			 (push (cons item perm) result))))
;; 	   result))))

;; (defun list->vector (lst)
;;   (let ((result (make-array (length lst) :adjustable t :fill-pointer 0)))
;;     (loop for item in lst
;; 	  do
;; 	  (vector-push-extend item result))
;;     result))

;; ;(list->vector '(1 2)) 

;; (defun vector-remove-if (pred vec)
;;   (let ((result (list)))
;;     (loop for item across vec
;; 	  do
;; 	  (if (not (funcall pred item))
;; 	      (push item result)))
;;     (list->vector (reverse result))))

;; (format t "building all-perms...~&")

;; (defun num-list->string (nlist)
;;   (let ((str ""))
;;     (loop for num in nlist
;; 	  do
;; 	  (setf str (concatenate 'string str (format nil "~d" num))))
;;     str))


;; (defun list> (left right)
;;   (cond ((or (null left)
;; 	     (null right))
;; 	 nil)
;; 	((= (first left)
;; 	    (first right))
;; 	 (list> (rest left)
;; 		   (rest right)))
;; 	((> (first left)
;; 	    (first right))
;; 	 t)
;; 	(t
;; 	 nil)))

;; (defun list< (left right)
;;   (cond ((or (null left)
;; 	     (null right))
;; 	 nil)
;; 	((= (first left)
;; 	    (first right))
;; 	 (list> (rest left)
;; 		   (rest right)))
;; 	((> (first left)
;; 	    (first right))
;; 	 nil)
;; 	(t
;; 	 t)))

;; ; (list< '(1 3 1) '(9 1 1))

;; (defvar *all-perms* (sort 
;; 		     (all-permutations '(0 1 2 3 4 5 6 7 8 9))
;; 		     #'list<))

;; ;(first *all-perms*)
;; ;(second *all-perms*)

;; ; (setf *all-perms* (mapcar #'num-list->string (all-permutations '(0 1 2 3 4 5 6 7 8 9))))
	  

;; (format t "all-perms:~a~&" (length *all-perms*))


;; ;(num-list->string '(1 2 3))

;; ;; this exhausts sbcl's heap
;; ;; (format t "building string-perms~&")
;; ;; (defvar *string-perms*
;; ;;   (mapcar #'num-list->string *all-perms*))
;; ;; (format t "string-perms:~a~&" (length *string-perms*))

;; '(loop for perm in *sorted-perms*
;;       for ii from 1 upto 1000000
;;       finally (return perm))

;; ;; (defun make-perm-spigot (lst)
;; ;;   (let ((item (first lst))
;; ;; 	(items (rest lst)))
;; ;;     #'(lambda ()
;; ;; 	(prog1
;; ;; 	    (cons item (all-permutations (remove-if (curry #'equalp item) items)))
;; ;; 	  (setf item (first items))
;; ;; 	  (setf items (rest items))))))

;; ;; (let ((perms (make-perm-spigot '(0 1 2 3 4 5 6 7 8 9))))
;; ;;   (format t "perms: ~a~&" (funcall perms))
;; ;;   (format t "perms: ~a~&" (funcall perms))
;; ;;   (format t "perms: ~a~&" (funcall perms)))


